/* Copyright 2019 The TensorFlow Authors. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
==============================================================================*/

package quake.quake_tensorflow_face_rec.detection.tflite;

import android.content.res.AssetFileDescriptor;
import android.content.res.AssetManager;
import android.graphics.Bitmap;
import android.graphics.RectF;
import android.os.Trace;
import android.util.Log;
import android.util.Pair;

import org.tensorflow.lite.Interpreter;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import quake.quake_tensorflow_face_rec.detection.env.Logger;


/**
 * Wrapper for frozen detection models trained using the Tensorflow Object Detection API:
 * - https://github.com/tensorflow/models/tree/master/research/object_detection
 * where you can find the training code.
 * <p>
 * To use pretrained models in the API or convert to TF Lite models, please see docs for details:
 * - https://github.com/tensorflow/models/blob/master/research/object_detection/g3doc/detection_model_zoo.md
 * - https://github.com/tensorflow/models/blob/master/research/object_detection/g3doc/running_on_mobile_tensorflowlite.md#running-our-model-on-android
 */
public class TFLiteObjectDetectionAPIModel
        implements SimilarityClassifier {

    private static final Logger LOGGER = new Logger();

    //private static final int OUTPUT_SIZE = 512;
    private static final int OUTPUT_SIZE = 192;

    // Only return this many results.
    private static final int NUM_DETECTIONS = 1;

    // Float model
    private static final float IMAGE_MEAN = 128.0f;
    private static final float IMAGE_STD = 128.0f;

    // Number of threads in the java app
    private static final int NUM_THREADS = 4;
    private boolean isModelQuantized;
    // Config values.
    private int inputSize;
    // Pre-allocated buffers.
    private Vector<String> labels = new Vector<String>();
    private int[] intValues;
    // outputLocations: array of shape [Batchsize, NUM_DETECTIONS,4]
    // contains the location of detected boxes
    private float[][][] outputLocations;
    // outputClasses: array of shape [Batchsize, NUM_DETECTIONS]
    // contains the classes of detected boxes
    private float[][] outputClasses;
    // outputScores: array of shape [Batchsize, NUM_DETECTIONS]
    // contains the scores of detected boxes
    private float[][] outputScores;
    // numDetections: array of shape [Batchsize]
    // contains the number of detected boxes
    private float[] numDetections;

    private float[][] embeedings;


    private ByteBuffer imgData;

    private Interpreter tfLite;

    // Face Mask Detector Output
    private float[][] output;

    //存放注册信息，包括姓名和描述所识别的内容
    private HashMap<String, Recognition> registered = new HashMap<>();
    private String mLog = "\n";

    //注册人脸
    public void register(String worker_id, Recognition rec) {
        registered.put(worker_id, rec);
        Log.d("hashmap", "registered=" + registered);

    }

    private TFLiteObjectDetectionAPIModel() {
    }

    /**
     * Memory-map the model file in Assets.
     */
    //导入内存映射资产中的模型文件
    private static MappedByteBuffer loadModelFile(AssetManager assets, String modelFilename) {
        FileInputStream inputStream = null;
        AssetFileDescriptor fileDescriptor = null;
        FileChannel fileChannel = null;
        try {
            fileDescriptor = assets.openFd(modelFilename);
            inputStream = new FileInputStream(fileDescriptor.getFileDescriptor());
            fileChannel = inputStream.getChannel();
            long startOffset = fileDescriptor.getStartOffset();
            long declaredLength = fileDescriptor.getDeclaredLength();
            return fileChannel.map(FileChannel.MapMode.READ_ONLY, startOffset, declaredLength);
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            if (inputStream != null) {
                try {
                    inputStream.close();
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
            if (fileDescriptor != null) {
                try {
                    fileDescriptor.close();
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
            if (fileChannel != null) {
                try {
                    fileChannel.close();
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }

        }

        return null;

    }

    /**
     * Initializes a native TensorFlow session for classifying images.
     * 初始化用于分类图像的本地TensorFlow会话
     *
     * @param assetManager  The asset manager to be used to load assets.
     *                      用于加载资产的资产管理器
     * @param modelFilename The filepath of the model GraphDef protocol buffer.
     *                      GraphDef协议缓冲区的模型的文件路径
     * @param labelFilename The filepath of label file for classes.
     *                      标签文件的文件路径
     * @param inputSize     The size of image input
     *                      图像输入的大小
     * @param isQuantized   Boolean representing model is quantized or not
     *                      布尔表示模型是否量化
     */
    public static SimilarityClassifier create(
            final AssetManager assetManager,
            final String modelFilename,
            final String labelFilename,
            final int inputSize,
            final boolean isQuantized)
            throws IOException {

        final TFLiteObjectDetectionAPIModel d = new TFLiteObjectDetectionAPIModel();


        String actualFilename = labelFilename.split("file:///android_asset/")[1];
        InputStream labelsInput = assetManager.open(actualFilename);
        BufferedReader br = new BufferedReader(new InputStreamReader(labelsInput));
        String line;
        while ((line = br.readLine()) != null) {
            LOGGER.w(line);
            d.labels.add(line);
        }
        br.close();

        d.inputSize = inputSize;

        try {
            d.tfLite = new Interpreter(loadModelFile(assetManager, modelFilename));
        } catch (Exception e) {
            throw new RuntimeException(e);
        }

        d.isModelQuantized = isQuantized;
        // Pre-allocate buffers.
        int numBytesPerChannel;
        if (isQuantized) {
            numBytesPerChannel = 1; // Quantized
        } else {
            numBytesPerChannel = 4; // Floating point
        }
        d.imgData = ByteBuffer.allocateDirect(1 * d.inputSize * d.inputSize * 3 * numBytesPerChannel);
        d.imgData.order(ByteOrder.nativeOrder());
        d.intValues = new int[d.inputSize * d.inputSize];

        d.tfLite.setNumThreads(NUM_THREADS);
        d.outputLocations = new float[1][NUM_DETECTIONS][4];
        d.outputClasses = new float[1][NUM_DETECTIONS];
        d.outputScores = new float[1][NUM_DETECTIONS];
        d.numDetections = new float[1];
        return d;
    }

    // looks for the nearest embeeding in the dataset (using L2 norm)
    // and retrurns the pair <id, distance>
    //在数据集中寻找最近的嵌入（使用L2范数）
    // 并返回<id,distance>对
    private Pair<String, Float> findNearest(float[] emb,int frameNum) {

        Pair<String, Float> ret = null;
        int num=0;
        mLog  = mLog+"---------第"+frameNum+"帧比对--------\n";
        for (Map.Entry<String, Recognition> entry : registered.entrySet()) {
            Log.d("findNearest", "registered:" + entry);
            Log.d("entry", "registered:" + entry.getValue());
            num++;
            final String name = entry.getKey();
            final float[] knownEmb = ((float[][]) entry.getValue().getExtra())[0];
            Log.d("findNearest", "getName:" + name);
            Log.d("findNearest", "getTitle:" + entry.getValue().getTitle());

            Log.d("findNearest", "knownEmb:" + knownEmb);
            float distance = 0;
            for (int i = 0; i < emb.length; i++) {
                float diff = emb[i] - knownEmb[i];
//              Log.d("diff","diff"+diff);
//          Log.d("findNearest","knownEmb:"+knownEmb[i]);
                distance += diff * diff;
            }
//      Log.d("knownEmb","|");
            distance = (float) Math.sqrt(distance);
//      Log.d("distance--",name+",distance:"+distance);
//      Log.d("recognizeResult",name+",distance:"+distance);
            mLog  = mLog+"和第"+num+"个人比对,ID:"+ name + ",distance:" + distance+"\n";
            Log.d("recognizeResult", "mLog" + mLog);

            if (ret == null || distance < ret.second) {
//          if(distance<=0.5){

                Log.e("recognizeResult", "和第"+num+"个人比对,ID:"+ name + ",distance:" + distance);

                ret = new Pair<>(name, distance);
                Log.d("recognizeResult", "找到了" + name);

//          }

            }
        }

        return ret;

    }


    //人脸识别
    //storeExtra标识是否注册人脸
    @Override
    public List<Recognition> recognizeImage(final Bitmap bitmap, String registerOrRecognize,int frameNum) {
        // Log this method so that it can be analyzed with systrace.
        //记录此方法，以便可以使用systrace对其进行分析
        Trace.beginSection("recognizeImage");

        Trace.beginSection("preprocessBitmap");
        // Preprocess the image data from 0-255 int to normalized float based
        // on the provided parameters.
        //根据提供的参数将图像数据从0-255int预处理为标准化浮点数
        //预处理输入图片，读取像素点，并将RGB三通道数值归一化，归一化后分布于-117~138
        Log.d("array", "intValues:" + intValues + ",bitmap.getWidth()" + bitmap.getWidth() + ",bitmap.getHeight():" + bitmap.getHeight());
        bitmap.getPixels(intValues, 0, bitmap.getWidth(), 0, 0, bitmap.getWidth(), bitmap.getHeight());

        imgData.rewind();
        for (int i = 0; i < inputSize; ++i) {
            for (int j = 0; j < inputSize; ++j) {
                int pixelValue = intValues[i * inputSize + j];
                if (isModelQuantized) {
                    // Quantized model
                    imgData.put((byte) ((pixelValue >> 16) & 0xFF));
                    imgData.put((byte) ((pixelValue >> 8) & 0xFF));
                    imgData.put((byte) (pixelValue & 0xFF));
                } else { // Float model
                    imgData.putFloat((((pixelValue >> 16) & 0xFF) - IMAGE_MEAN) / IMAGE_STD);//归一化通道R
                    imgData.putFloat((((pixelValue >> 8) & 0xFF) - IMAGE_MEAN) / IMAGE_STD);//归一化通道G
                    imgData.putFloat(((pixelValue & 0xFF) - IMAGE_MEAN) / IMAGE_STD);//归一化通道B
                }
            }
        }
        Trace.endSection(); // preprocessBitmap

        // Copy the input data into TensorFlow.
        //将输入数据填充到TensorFlow中，并Feed数据给模型
        //
        Trace.beginSection("feed");

//==========================================以上都是预处理图像
        Object[] inputArray = {imgData};

        Trace.endSection();

// Here outputMap is changed to fit the Face Mask detector
        Map<Integer, Object> outputMap = new HashMap<>();

        embeedings = new float[1][OUTPUT_SIZE];
        outputMap.put(0, embeedings);


        // Run the inference call.
        Trace.beginSection("run");
        //tfLite.runForMultipleInputsOutputs(inputArray, outputMapBack);
        //运行模型
        tfLite.runForMultipleInputsOutputs(inputArray, outputMap);
        Trace.endSection();

        String res = "[";
        for (int i = 0; i < embeedings[0].length; i++) {
            res += embeedings[0][i];
//            if (i < embeedings[0].length - 1) res += ",\n ";
            if (i < embeedings[0].length - 1) res += ",";
        }
        res += "]";
        Log.d("recognizeImage", "embedding的个数：" + embeedings[0].length);
        Log.d("recognizeImage", "res=" + res);


        float distance = Float.MAX_VALUE;
        String id = "0000";
        String label = "?";

        //如果有注册的人脸
        Log.d("???", "registered.size()" + registered.size());
        Log.d("???", "registered.size()" + registered);

        Log.d("???", "registerOrRecognize" + registerOrRecognize);

        if (registered.size() > 0 && registerOrRecognize.equals("recognize")) {
            //LOGGER.i("dataset SIZE: " + registered.size());
            //寻找最接近的数据

            final Pair<String, Float> nearest = findNearest(embeedings[0],frameNum);
            Log.d("findNearest", "nearest: " + nearest.first + " - distance: " + nearest.second);
            if (nearest != null) {

                final String name = nearest.first;
                label = name;
                distance = nearest.second;

                LOGGER.i("nearest: " + name + " - distance: " + distance);
                Log.d("wdnmd", "nearest: " + name + " - distance: " + distance);

            }
        }


        final int numDetectionsOutput = 1;
        final ArrayList<Recognition> recognitions = new ArrayList<>(numDetectionsOutput);
        Recognition rec = new Recognition(
                id,
                label,
                distance,
                new RectF());
        Log.d("wdnmd", " distance: " + distance);
        recognitions.add(rec);
//    Log.d("recognizeImage","storeExtra:"+storeExtra);
        //storeExtra判断是否需要注册人脸
//    if (storeExtra) {
        rec.setExtra(embeedings);
        rec.setEmbeddings(embeedings);
        rec.setLOG(mLog);
//    }
        Log.d("recognizeResult", "rec.getLOG()" + rec.getLOG());

        Trace.endSection();
        return recognitions;
    }

    @Override
    public void enableStatLogging(final boolean logStats) {
    }

    @Override
    public String getStatString() {
        return "";
    }

    @Override
    public void close() {
    }

    public void setNumThreads(int num_threads) {
        if (tfLite != null) tfLite.setNumThreads(num_threads);
    }

    @Override
    public void setUseNNAPI(boolean isChecked) {
        if (tfLite != null) tfLite.setUseNNAPI(isChecked);
    }
}
